"
 COPYRIGHT (c) 1997 by Claus Gittinger / eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"




View subclass:#DSVColumnView
	instanceVariableNames:'editValue editView selectedColIndex selectedRowIndex rowHeight
		columnDescriptors viewOrigin registererImages list fgColor canFit
		listSizeChanged beDependentOfRows has3Dsepartors bgColor
		hgLgFgColor hgLgBgColor actionBlock doubleClickActionBlock
		separatorLightColor separatorDarkColor verticalSpacing
		horizontalSpacing'
	classVariableNames:'DefaultForegroundColor DefaultBackgroundColor
		DefaultHilightForegroundColor DefaultHilightBackgroundColor
		DefaultSeparatorDarkColor DefaultSeparatorLightColor'
	poolDictionaries:''
	category:'Views-DataSet'
!

!DSVColumnView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1997 by Claus Gittinger / eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"



!

documentation
"
    implements a scrollable selection view based on rows and columns

    [Instance variables:]

        editValue               <Model>                 current editing model
        editView                <View>                  current editing component
        selectedColIndex        <Integer>               selected column index or 0
        selectedRowIndex        <Integer>               selected row index or 0
        rowHeight               <Integer>               maximum height of any row
        columnDescriptors       <SequancableCollection> list of column descriptors
        list                    <SequancableCollection> list of rows
        fgColor                 <Color>                 foreground color
        registererImages        <IdentityDictionary>    list of images registered on the device
        beDependentOfRows       <Coolean>               keep rows dependent; on default is disabled.
                                                        in case of enabled a row can raise a change
                                                        notification whithout a parameter which
                                                        will force a redraw of the row or the
                                                        readSelector of the column which will
                                                        redraw the cell in the row only.

    [author:]
        Claus Atzkern

    [see also:]
        DataSetColumnSpec
        DataSetColumn
        DataSetView
"

! !

!DSVColumnView class methodsFor:'defaults'!

horizontalSpacing
    ^ 4

!

updateStyleCache
    "extract values from the styleSheet and cache them in class variables
    "
    <resource: #style (#textForegroundColor #'scrollableView.backgroundColor'
                       #'selection.hilightForegroundColor'
                       #'selection.hilightBackgroundColor'   )>

    DefaultForegroundColor        := StyleSheet colorAt:'textForegroundColor' default:(Color black).
    DefaultBackgroundColor        := StyleSheet at:'scrollableView.backgroundColor' default:DefaultViewBackgroundColor.

    DefaultHilightForegroundColor := StyleSheet colorAt:'selection.hilightForegroundColor' default:DefaultForegroundColor.
    DefaultHilightBackgroundColor := StyleSheet colorAt:'selection.hilightBackgroundColor' default:(Color veryLightGrey).
    DefaultSeparatorDarkColor     := DefaultShadowColor ? (Color grayPercent:40).
    DefaultSeparatorLightColor    := DefaultLightColor  ? (Color grayPercent:75).
"
self updateStyleCache.
"

    "Modified: / 26.10.1997 / 17:09:07 / cg"
!

verticalSpacing
    ^ 2

! !

!DSVColumnView methodsFor:'accessing'!

action:aOneArgAction
    actionBlock := aOneArgAction


!

doubleClickAction:aOneArgAction
    doubleClickActionBlock := aOneArgAction


! !

!DSVColumnView methodsFor:'accessing behavior'!

beDependentOfRows
    "make myself dependent of any row; in this case any change notification
     raised by a row is catched and the cell identified by the 'readSelector'
     is redrawn. In case of a nil readSelector, the whole raw is redrawn.
        -> row changed:'what'
     On default the attribute is set to true (enabled).
    "
    ^ beDependentOfRows
!

beDependentOfRows:aBool
    "make myself dependent of any row; in this case any change notification
     raised by a row is catched and the cell identified by the 'readSelector'
     is redrawn. In case of a nil readSelector, the whole raw is redrawn.
        -> row changed:'what'
     On default the attribute is set to true (enabled).
    "
    aBool ~~ beDependentOfRows ifTrue:[
        (beDependentOfRows := aBool) ifTrue:[
            self do:[:aRow| aRow addDependent:self]
        ] ifFalse:[
            self do:[:aRow| aRow removeDependent:self]
        ]
    ].
! !

!DSVColumnView methodsFor:'accessing columns'!

columnAt:anIndex
    "get the column at an index
    "
    ^ columnDescriptors at:anIndex
!

columnDescriptors:colDesc
    "set the columnDescriptors
    "
    self setSelectColIndex:0 rowIndex:0.

    colDesc size ~~ 0 ifTrue:[
        columnDescriptors := colDesc collect:[:aCol|
            DataSetColumn on:self description:aCol
        ]
    ] ifFalse:[
        columnDescriptors := #()
    ].
    canFit := (columnDescriptors findFirst:[:aCol|aCol canResize]) ~~ 0.
    preferredExtent := nil.

    listSizeChanged := true.
    self preferredExtentChanged.

    realized ifTrue:[
        self fitColumns ifFalse:[
            self invalidate.
            self contentsChanged.
        ]
    ].
    self changed:#sizeOfColumns.
    listSizeChanged := false.


!

firstColumn
    "get the first column
    "
    ^ columnDescriptors at:1

!

lastColumn
    "get the last column
    "
    ^ columnDescriptors last

! !

!DSVColumnView methodsFor:'accessing rows'!

at:aRowNr
    "return the row at an index, aRowNr
    "
    ^ list at:aRowNr
!

at:aRowNr ifAbsent:exceptionBlock
    "return the element at a aRowNr. If the index is invalid, return the
     result of evaluating the exceptionblock
    "
    ^ list at:aRowNr ifAbsent:exceptionBlock
!

at:aRowNr put:aRow
    "change the row at an index. The added row is returned
    "
    beDependentOfRows ifTrue:[
        (list at:aRowNr) removeDependent:self.
        aRow addDependent:self
        
    ] ifFalse:[
        list at:aRowNr put:aRow
    ].
    self redrawRowAt:aRowNr.
    ^ aRow.
!

first
    "return the first row
    "
    ^ list at:1
!

last
    "return the last row
    "
    ^ list last
!

list
    "get the list of rows
    "
    ^ list



!

list:aList
    "set the list of rows
    "
    |oldSize|

    self setSelectColIndex:0 rowIndex:0.

    aList size ~~ 0 ifTrue:[  "/ otherwise keep old values
        self columnsDo:[:aCol| aCol invalidate ].
        preferredExtent := nil.
    ].
    oldSize := list size.

    beDependentOfRows ifTrue:[
        self do:[:aRow| aRow removeDependent:self].
        list := aList.
        self do:[:aRow| aRow addDependent:self]
    ] ifFalse:[
        list := aList
    ].
    list size == 0 ifTrue:[
        list := nil
    ].
    self preferredExtent.
    self recomputeHeightOfContents.

    realized ifTrue:[
        list size ~~ 0 ifTrue:[
            self fitColumns ifTrue:[^ self ].
        ] ifFalse:[
            oldSize == 0 ifTrue:[
                ^ self
            ]
        ].
        self redraw.
    ]



! !

!DSVColumnView methodsFor:'accessing visibility'!

backgroundColor
    "get the background color of the rows
    "
    ^ bgColor


!

backgroundColor:aColor
    "set the background color of the rows
    "
    bgColor ~~ aColor ifTrue:[
        super viewBackground:bgColor.

        self realized ifTrue:[
            bgColor := aColor on:device.
            self invalidate
        ] ifFalse:[
            bgColor := aColor
        ].
        self changed:#style.
    ]
!

font:aFont
    "set the font for all shown rows.
    "
    (aFont notNil and:[aFont ~~ font]) ifTrue:[
        super font:(aFont on:device).
        self columnsDo:[:aCol| aCol invalidate ].
        self preferredExtentChanged.

        realized ifTrue:[
            self invalidate.
            self contentsChanged
        ].
        self changed:#style.
    ]

!

foregroundColor
    "return the foreground color of the rows
    "
    ^ fgColor

!

foregroundColor:aColor
    "set the foreground color of the rows
    "
    fgColor ~~ aColor ifTrue:[
        self realized ifTrue:[
            fgColor := aColor on:device.
            self invalidate
        ] ifFalse:[
            fgColor := aColor
        ].
        self changed:#style.
    ]

!

has3Dsepartors
    ^ has3Dsepartors
!

has3Dsepartors:aBool

    aBool ~~ has3Dsepartors ifTrue:[
        has3Dsepartors := aBool.
        self columnsDo:[:aCol| aCol invalidate ].
        self preferredExtentChanged.

        realized ifTrue:[
            self invalidate.
            self contentsChanged
        ].
        self changed:#style.
    ]


!

hgLgBgColor
    ^ hgLgBgColor
!

hgLgFgColor
    ^ hgLgFgColor
!

horizontalSpacing
    "horizontal spacing used by columns
    "
    ^ horizontalSpacing
!

horizontalSpacing:aNumber
    "horizontal spacing used by columns
    "
    horizontalSpacing ~~ aNumber ifTrue:[
        horizontalSpacing := aNumber.
        self preferredExtentChanged.
    ].
!

separatorDarkColor
    ^ separatorDarkColor


!

separatorLightColor
    ^ separatorLightColor


!

verticalSpacing
    "vertical spacing used by columns
    "
    ^ verticalSpacing
!

verticalSpacing:aNumber
    "vertical spacing used by columns
    "
    verticalSpacing ~~ aNumber ifTrue:[
        verticalSpacing := aNumber.
        self preferredExtentChanged.
    ].
! !

!DSVColumnView methodsFor:'adding & removing rows'!

add:aRow
    "insert row at the end of the list of rows
    "
    ^ self add:aRow beforeIndex:(1 + list size)
!

add:aRow afterIndex:aRowNr
    "add a new row after slot aRowNr and redisplay; returns nil in case
     of an invalid index index or the row added to the list
    "
    ^ self add:aRow beforeIndex:(aRowNr + 1)
!

add:aRow beforeIndex:aRowNr
    "add a new row before slot aRowNr and redisplay; returns nil in case
     of an invalid index index or the row added to the list
    "
    |y0 y1 h dH|

    list isNil ifTrue:[
        list := OrderedCollection new.
        preferredExtent := nil.         "/ force a recomputation
    ] ifFalse:[
        list isArray ifTrue:[list := list asOrderedCollection]
    ].
    list add:aRow beforeIndex:aRowNr.
    self recomputeHeightOfContents.

    beDependentOfRows ifTrue:[
        aRow addDependent:self.
    ].
    (aRowNr <= selectedRowIndex) ifTrue:[
        selectedRowIndex := selectedRowIndex + 1.
    ].
    y0 := (aRowNr - 1) * rowHeight.

    y0 < viewOrigin y ifTrue:[
        self originWillChange.
        viewOrigin y:(rowHeight + viewOrigin y).
        self originChanged:(0 @ rowHeight).        
    ].
    shown ifTrue:[
        y0 := self yVisibleOfRowNr:aRowNr.
        y1 := y0 + rowHeight.

        (y1 > margin and:[y0 < (height - margin)]) ifTrue:[
            h := height - margin - y1.
            y0 := y0 max:margin.
            dH := y1 - y0.
            self catchExpose.

            self copyFrom:self x:0 y:y0
                             toX:0 y:y1
                           width:width height:h async:true.
            self waitForExpose.
            self redrawX:margin y:y0 width:width height:dH.
        ]
    ].
    listSizeChanged := true.
    self contentsChanged.
    listSizeChanged := false.
    ^ aRow
!

addAll:aCollection beforeIndex:aRowNr
    "add a collection of rows before slot aRowNr and redisplay
    "
    aCollection size ~~ 0 ifTrue:[
        list size == 0 ifTrue:[
            ^ self list:aCollection
        ].
        aCollection reverseDo:[:aRow|
            self add:aRow beforeIndex:aRowNr
        ]
    ].
!

addFirst:aRow
    "insert row at the start of the list of rows
    "
    ^ self add:aRow beforeIndex:1
!

removeFirst
    "remove first row; returns the removed row
    "
    self removeIndex:1
!

removeIndex:aRowNr
    "remove row at an index; returns the removed row
    "
    |y0 y1 h dY oY yB noRedraw row|

    row := list at:aRowNr ifAbsent:[^ self subscriptBoundsError:aRowNr].

    list isArray ifTrue:[
        list := list asOrderedCollection.
    ].
    selectedRowIndex == aRowNr ifTrue:[
        self selectColIndex:0 rowIndex:0
    ] ifFalse:[
        selectedRowIndex > aRowNr ifTrue:[
            selectedRowIndex := selectedRowIndex - 1
        ]
    ].
    y1 := aRowNr * rowHeight.
    y0 := y1 - rowHeight.
    yB := y1 + margin - viewOrigin y.

    list removeIndex:aRowNr.
    self recomputeHeightOfContents.

    y0 < (oY := viewOrigin y) ifTrue:[
        (noRedraw := y1 <= oY) ifFalse:[dY := y0 - oY]
                                ifTrue:[dY := rowHeight negated].
        self originWillChange.
        viewOrigin y:(dY + oY).
        self originChanged:(0 @ dY).        
    ] ifFalse:[
        noRedraw := y0 > (height + viewOrigin y)
    ].

    (noRedraw or:[shown not]) ifFalse:[
        y1 := yB.
        y0 := self yVisibleOfRowNr:aRowNr.
        h  := height - margin - yB.
        y0 := y0 max:margin.

        h > 0 ifTrue:[
            self catchExpose.
            self copyFrom:self x:0 y:yB toX:0 y:y0 width:width height:h async:true.
            self waitForExpose.
        ].
        y0 := y0 + h.
        self redrawX:margin y:y0 width:width height:(height - y0).
    ].
    listSizeChanged := true.
    self contentsChanged.
    listSizeChanged := false.

    beDependentOfRows ifTrue:[
        row removeDependent:self
    ].

    ^ row.
!

removeLast
    "remove last row; the row is returned
    "
    ^ self removeIndex:list size
! !

!DSVColumnView methodsFor:'change & update'!

update:something with:aParameter from:changedObject
    "one of my rows/cells changed its value
    "
    |start colNr rowNr|

    (start := self indexOfFirstRowShown) ~~ 0 ifTrue:[
        rowNr := list identityIndexOf:something
                           startingAt:start
                             endingAt:(start + (height // rowHeight)).

        rowNr ~~ 0 ifTrue:[
            something notNil ifTrue:[
                colNr := columnDescriptors findFirst:[:aCol|
                    aCol readSelector == something
                ]
            ] ifFalse:[
                colNr := 0
            ].
            self redrawRowAt:rowNr colAt:colNr                
        ]
    ]
! !

!DSVColumnView methodsFor:'enumerating columns'!

columnsDo:aOneArgBlock
    "evaluate the argument, aOneArgBlock for every column
    "
    columnDescriptors do:aOneArgBlock


!

columnsFrom:start to:stop do:aOneArgBlock
    "evaluate the argument, aOneArgBlock for the columns with index start to
     stop in the collection of column descriptors
    "
    columnDescriptors from:start to:stop do:aOneArgBlock


! !

!DSVColumnView methodsFor:'enumerating rows'!

do:aOneArgBlock
    "evaluate the argument, aOneArgBlock for every row
    "
    list size ~~ 0 ifTrue:[list do:aOneArgBlock]

!

findFirst:aOneArgBlock
    "find the first row, for which evaluation of the argument, aOneArgBlock
     returns true; return its index or 0 if none detected.
    "
    list size ~~ 0 ifTrue:[
        ^ list findFirst:aOneArgBlock
    ].
    ^ 0

!

from:start to:stop do:aOneArgBlock
    "evaluate the argument, aOneArgBlock for the rows with index start to
     stop in the collection of rows
    "
    list size >= start ifTrue:[
        list from:start to:stop do:aOneArgBlock
    ]
! !

!DSVColumnView methodsFor:'event handling'!

buttonMultiPress:button x:x y:y
    "a button was pressed twice - handle doubleclick here
    "
    ((button == 1) or:[button == #select]) ifFalse:[
        ^ super buttonMultiPress:button x:x y:y
    ].
    selectedRowIndex == (self yVisibleToRowNr:y) ifTrue:[
        (     selectedColIndex == 0
          or:[selectedColIndex == (self xVisibleToColNr:x)]
        ) ifTrue:[
            self doubleClicked
        ]
    ]




!

buttonPress:button x:x y:y
    "a button was pressed - handle selection here
    "
    |rowNr colNr menu col|

    (     (rowNr := self yVisibleToRowNr:y) notNil
     and:[(colNr := self xVisibleToColNr:x) notNil]
    ) ifTrue:[
        ((button == 2) or:[button == #menu]) ifTrue:[           "/ menu for cell
            selectedRowIndex == rowNr ifTrue:[

                (selectedColIndex == 0) ifTrue:[                "/ row selected
                    col := columnDescriptors detect:[:c| c rendererType == #rowSelector]
                                             ifNone:[nil].

                    col notNil ifTrue:[
                        menu := col menuForRow:(self selectedRow).
                        menu notNil ifTrue:[
                            ^ menu startUp
                        ]
                    ].
                    ^ self
                ].
                selectedColIndex == colNr ifTrue:[
                    menu := self selectedColumn menuForRow:(self selectedRow).
                    menu notNil ifTrue:[
                        ^ menu startUp
                    ].
                ].
            ].
        ] ifFalse:[
            ((button == 1) or:[button == #select]) ifTrue:[     "/ selection change
                ^ self selectColIndex:colNr rowIndex:rowNr
            ]
        ]
    ].
    ^ super buttonPress:button x:x y:y


!

contentsChanged
    "contents changed - move origin up if possible
    "
    |y|

    shown ifTrue:[
        self recomputeHeightOfContents.
        y := self maxViewOriginY.

        viewOrigin y > y ifTrue:[
            self scrollTo:(viewOrigin x @ y)
        ] ifFalse:[
            self updateEditViewOrigin.
        ]
    ].
    super contentsChanged


!

doubleClicked
    "handle a double click
    "
    (doubleClickActionBlock notNil and:[self hasSelection]) ifTrue:[
        doubleClickActionBlock value:selectedRowIndex
    ].
!

keyPress:key x:x y:y
    "a key was pressed - handle page-keys here
    "
    <resource: #keyboard (#PreviousPage #NextPage #HalfPageUp #HalfPageDown
                          #BeginOfText #EndOfText
                          #ScrollUp #ScrollDown
                          #CursorUp #CursorDown #CursorRight #CursorLeft)>

    |sensor n max idx col|

    (key == #PreviousPage) ifTrue: [^ self pageUp].
    (key == #NextPage)     ifTrue: [^ self pageDown].
    (key == #HalfPageUp)   ifTrue: [^ self halfPageUp].
    (key == #HalfPageDown) ifTrue: [^ self halfPageDown].

    (key == #BeginOfText) ifTrue:[^ self scrollToTop].
    (key == #EndOfText) ifTrue:[^ self scrollToBottom].

    sensor := self sensor.

    (key == #ScrollUp) ifTrue:[
        sensor isNil ifTrue:[
            n := 1
        ] ifFalse:[
            n := 1 + (sensor compressKeyPressEventsWithKey:#ScrollUp).
        ].
        ^ self scrollUp:(n * self verticalScrollStep)
    ].
    (key == #ScrollDown) ifTrue:[
        sensor isNil ifTrue:[
            n := 1
        ] ifFalse:[
            n := 1 + (sensor compressKeyPressEventsWithKey:#ScrollDown).
        ].
        ^ self scrollDown:(n * self verticalScrollStep)
    ].
    self hasSelection ifFalse:[^ self].

    key == #Return ifTrue:[
        ^ self doubleClicked
    ].

    (key == #CursorUp or:[key == #CursorDown]) ifTrue:[
        sensor isNil ifTrue:[
            n := 1
        ] ifFalse:[
            n := 1 + (sensor compressKeyPressEventsWithKey:key).
        ].
        max := self numberOfRows.
        (n := n \\ max) == 0 ifTrue:[^ self ].

        key == #CursorUp ifTrue:[
            (n := selectedRowIndex - n) <= 0  ifTrue:[n := max + n]
        ] ifFalse:[
            (n := selectedRowIndex + n) > max ifTrue:[n := n - max]
        ].
        ^ self selectColIndex:selectedColIndex rowIndex:n.
    ].

    (key == #CursorRight or:[key == #CursorLeft]) ifTrue:[
        selectedColIndex == 0 ifTrue:[                          "/ line selected
            ^ self
        ].
        sensor isNil ifTrue:[
            n := 1
        ] ifFalse:[
            n := 1 + (sensor compressKeyPressEventsWithKey:key).
        ].
        (n := n \\ (self numberOfSelectableColumns)) == 0 ifTrue:[^ self ].

        idx := selectedColIndex.
        max := self numberOfColumns.

        key == #CursorLeft ifTrue:[
            [n ~~ 0] whileTrue:[
                (idx := idx - 1) == 0 ifTrue:[idx := max].
                (self columnAt:idx) canSelect ifTrue:[n := n - 1]
            ]
        ] ifFalse:[
            [n ~~ 0] whileTrue:[
                idx == max ifTrue:[idx := 1]
                              ifFalse:[idx := idx + 1].
                (self columnAt:idx) canSelect ifTrue:[n := n - 1]
            ]
        ].

        ^ self selectColIndex:idx rowIndex:selectedRowIndex.
    ].
    key isCharacter ifFalse:[
        ^ self
    ].

    col := self columnAt:selectedColIndex.

    "/ search forward
    idx := col findRowNrStartingWithChar:key
                                   start:(selectedRowIndex + 1)
                                    stop:(self numberOfRows).

    idx == 0 ifTrue:[
        "/ search from begin
        idx := col findRowNrStartingWithChar:key
                                       start:1
                                        stop:(selectedRowIndex - 1).
    ].
    idx ~~ 0 ifTrue:[
        self selectColIndex:selectedColIndex rowIndex:idx.
    ].





!

originChanged:delta
    "this one is sent, after the origin of my contents has changed -
     tell dependents (i.e. scrollers) about this
    "
    super originChanged:delta.
    self updateEditViewOrigin.
!

sizeChanged:how
    "size changed - move origin up if possible
    "
    super sizeChanged:how.

    (shown and:[self numberOfColumns ~~ 0 and:[self fitColumns not]]) ifTrue:[
        self scrollTo:viewOrigin.       "/ validates viewOrigin
        self invalidate.
        self scrollToSelection
    ]

! !

!DSVColumnView methodsFor:'gc operations'!

registerImage:anImage key:aKey
    "any row can register an image with a unique identifier a key symbol
    "
    |img|

    (img := registererImages at:aKey ifAbsent:nil) notNil ifTrue:[
        ^ img
    ].
    img := anImage onDevice:device.
    img clearMaskedPixels.
    registererImages at:aKey put:img.
    ^ img
!

registeredImageAt:aKey
    "any row can register an image with a unique identifier
    "
    ^ registererImages at:aKey ifAbsent:nil
!

releaseAllRegisteredImages
    "release all registered images
    "
    registererImages := IdentityDictionary new.
! !

!DSVColumnView methodsFor:'initialization'!

create
    "set foreground color
    "
    super create.
    fgColor     := fgColor     on:device.
    bgColor     := bgColor     on:device.
    hgLgFgColor := hgLgFgColor on:device.
    hgLgBgColor := hgLgBgColor on:device.
    separatorLightColor := separatorLightColor on:device.
    separatorDarkColor  := separatorDarkColor  on:device.

!

destroy
    "remove dependencies
    "
    beDependentOfRows ifTrue:[
        self do:[:aRow| aRow removeDependent:self].
    ].
    super destroy
!

initStyle
    "setup viewStyle specifics"

    super initStyle.

    DefaultForegroundColor isNil ifTrue:[
        self class updateStyleCache
    ].
    fgColor     := DefaultForegroundColor.
    bgColor     := DefaultBackgroundColor.
    hgLgFgColor := DefaultHilightForegroundColor.
    hgLgBgColor := DefaultHilightBackgroundColor.
    separatorDarkColor  := DefaultSeparatorDarkColor.
    separatorLightColor := DefaultSeparatorLightColor.
!

initialize
    "set default attributes
    "
    super initialize.

    viewOrigin := 0@0.
    font       := font on:device.
    "/ self bitGravity:#NorthWest.
    rowHeight  := font height.
    selectedRowIndex := 0.
    selectedColIndex := 0.
    registererImages := IdentityDictionary new.
    has3Dsepartors   := false.
    columnDescriptors := #().
    beDependentOfRows := false.
    listSizeChanged   := false.
    verticalSpacing   := self class verticalSpacing.
    horizontalSpacing := self class horizontalSpacing.
!

realize
    "recompute contents; set selection
    "
    |rowNr colNr|

    self  recomputeHeightOfContents.
    super realize.

    rowNr := selectedRowIndex.
    colNr := selectedColIndex.
    selectedColIndex := 0.
    selectedRowIndex := 0.
    self fitColumns.
    self setSelectColIndex:colNr rowIndex:rowNr.
    self bitGravity:#NorthWest.

! !

!DSVColumnView methodsFor:'private'!

fitColumns
    "fit columns to view
    "
    |lastColumn dX sz expand resizables deltaWidth|

    self canFit ifFalse:[^ false ].
    resizables := 0.
    self preferredExtent.

    self columnsDo:[:aCol|
        aCol canResize ifTrue:[
            resizables := resizables + 1.
            deltaWidth := aCol setMinWidth.
            lastColumn := aCol.
        ]
    ].
    resizables == 0 ifTrue:[ ^ false ].

    sz := self widthOfContents.

    (expand := sz < width) ifTrue:[
        deltaWidth := width - sz + margin + margin.
    ] ifFalse:[
        deltaWidth == 0 ifTrue:[ ^ false ].
    ].

    expand ifTrue:[
        (dX := deltaWidth // resizables) ~~ 0 ifTrue:[
            self columnsDo:[:aCol|aCol canResize ifTrue:[aCol growWidth:dX]]
        ].
        lastColumn growWidth:(deltaWidth - (dX * resizables)).
    ].
    self preferredExtentChanged.
    self invalidate.

    self hasSelection ifTrue:[
        editView notNil ifTrue:[
            editView width:(self selectedColumn width - 1 - self separatorSize)
        ].
        self scrollToSelection.
    ].
    self contentsChanged.
  ^ true


!

maxViewOriginY
    "returns the maximum possible y of the view origin
    "
    |y|

    y := self heightOfContents - self innerHeight.
  ^ y max:0

!

updateEditViewOrigin
    "update origin of the editView
    "
    |x y|

    editView notNil ifTrue:[
        y := 1 + (self yVisibleOfRowNr:selectedRowIndex).
        x := 1 + (self xVisibleOfColNr:selectedColIndex).

        editView origin:(x @ y).
    ].

!

xVisibleOfColNr:aColNr
    "returns visible x assigned to a column number
    "
    |x|

    x := margin - viewOrigin x.
    self columnsFrom:1 to:(aColNr - 1) do:[:aCol| x := x + aCol width ].
  ^ x

!

xVisibleToColNr:x
    |x0 nr|

    x0 := x + viewOrigin x - margin.
    nr := 1.

    self columnsDo:[:aCol|
        (x0 := x0 - aCol width) <= 0 ifTrue:[^ nr].
        nr := nr + 1.
    ].
    ^ nil.

!

yVisibleOfRowNr:aRowNr
    "returns visible y assigned to the row number
    "
    ^ (aRowNr - 1) * rowHeight + margin - viewOrigin y

!

yVisibleToRowNr:y
    "returns row number or nil
    "
    |y0|

    y0 := (y + viewOrigin y - margin) // rowHeight + 1.
  ^ (y0 <= self numberOfRows) ifTrue:[y0] ifFalse:[nil]

! !

!DSVColumnView methodsFor:'queries'!

canFit
    "returns true if columns can be fit to view
    "
    ^ canFit
!

indexOfFirstRowShown
    "returns index of first row shown
    "
    |idx|

    idx := (self yOriginOfContents // rowHeight) + 1.
  ^ (idx <= list size) ifTrue:[idx] ifFalse:[0]

!

listSizeChanged
    ^ listSizeChanged
!

numberOfColumns
    "returns number of columns
    "
    ^ columnDescriptors size

!

numberOfRows
    "returns number of raws
    "
    ^ list size

!

numberOfSelectableColumns
    "returns number of selectable columns
    "
    |num|

    num := 0.
    self columnsDo:[:aCol| aCol canSelect ifTrue:[num := num + 1] ].
    ^ num
!

rowHeight
    "get the height of the highest row in pixels
    "
    ^ rowHeight

!

separatorSize
    "returns vertical/horizontal size of a separator dependent on the
     3D effect.
    "
    ^ has3Dsepartors ifTrue:[4] ifFalse:[1]

!

size
    "returns number of raws
    "
    ^ list size

! !

!DSVColumnView methodsFor:'recomputation'!

preferredExtent
    "recompute preferred extent
    "
    |x|

    preferredExtent notNil ifTrue:[
        ^ preferredExtent
    ].
    x := 3.
    rowHeight := 0.

    self columnsDo:[:aCol|
        rowHeight := (aCol heightOfHighestRow) max:rowHeight.
        x := x + (aCol minWidth).
    ].
    rowHeight := rowHeight + (2 * verticalSpacing) + (self separatorSize).
    preferredExtent := (x + margin + margin) @ (self numberOfRows * rowHeight).

    x := (rowHeight - font height) // 2 + font ascent.
    self columnsDo:[:aCol|aCol textInsetChanged:x].
  ^ preferredExtent


!

preferredExtentChanged
    |y x|

    y := viewOrigin y.
    x := viewOrigin x.

    (y ~~ 0 or:[x ~~ 0]) ifTrue:[
        self originWillChange.
        viewOrigin := 0 @ 0.
        preferredExtent := nil.
        self originChanged:(x negated  @ y negated).
    ]
!

recomputeHeightOfContents

    preferredExtent notNil ifTrue:[
        preferredExtent y:(rowHeight * self numberOfRows)
    ] ifFalse:[
        self preferredExtent
    ].
! !

!DSVColumnView methodsFor:'redrawing'!

forceRedraw
    "a redraw forced by any other component
    "
    self redraw
!

invalidate
    "recompute extent before repair range
    "
    self  recomputeHeightOfContents.
    super invalidate.


!

redraw
    "redraw complete view
    "
    self redrawX:0 y:0 width:width height:height.

!

redrawRowAt:aRowNr
    "redraw total row at an index
    "
    self redrawRowAt:aRowNr colAt:0
!

redrawRowAt:aRowNr colAt:aColNr
    "redraw visible row in case of aColNr == 0 or the column in a row;
     in case that the row/column is hidden no redraw is done
    "
    |x y h w|

    (shown and:[aRowNr between:1 and:self numberOfRows]) ifTrue:[
        h := rowHeight - 1.

        (y := self yVisibleOfRowNr:aRowNr) < margin ifTrue:[
            (h := h + y) <= margin ifTrue:[
                ^ self
            ].                                                  "/ row not visible
            h := h - margin.
            y := margin.
        ] ifFalse:[
            y >= height ifTrue:[^ self].
        ].
        aColNr ~~ 0 ifTrue:[                                    "/ redraw column in row
            w := (self columnAt:aColNr) width - 1.

            (x := self xVisibleOfColNr:aColNr) < margin ifTrue:[
                (w := w + x) <= margin ifTrue:[
                    ^ self
                ].                                              "/ column not visible
                w := w - margin.
                x := margin.
            ] ifFalse:[
                x >= width ifTrue:[^ self].
            ]
        ] ifFalse:[                                             "/ redraw whole row
            x := margin.
            w := width.
        ].
        self redrawX:x y:y width:w height:h
    ]
!

redrawX:x y:y width:w height:h
    "redraw part of myself immediately, given logical coordinates 
    "
    |x0 x1 c0 c1 yT xM yTop yBot start stop prevTrans transPoint prevClipArea clipHeight|

    shown ifFalse:[^ self].

    self paint:bgColor.
    self fillRectangleX:x y:y width:w height:h.

    yT := margin - viewOrigin y.
    c0 := y - yT.

    start := (c0 // rowHeight) + 1.
    stop  := (c0 + h - 1 // rowHeight + 1) min:(self numberOfRows).

    stop < start ifTrue:[
        ^ self
    ].
    xM             := (x + w) min:(width - margin).
    prevTrans      := transformation.
    prevClipArea   := clipRect.
    clipRect       := nil.
    transformation := WindowingTransformation new.
    clipHeight     := height - margin - margin.
    transPoint     := Point x:0 y:yT.

    x0    := margin - (viewOrigin x).
    yTop  := (start - 1) * rowHeight.
    yBot  := stop * rowHeight.

    self columnsDo:[:aCol|
        ((x1 := x0 + aCol width) >= x and:[x0 < xM]) ifTrue:[
            c0 := x0 max:margin.
            c1 := (x1 + 1) min:xM.

            device setClipX:c0 y:margin width:(c1 - c0) height:clipHeight
                         in:drawableId gc:gcId.

            transPoint x:x0.
            transformation translation:transPoint.
            aCol drawFrom:start to:stop yTop:yTop yBot:yBot with:fgColor and:bgColor.
        ].
        x0 := x1
    ].
    transformation := prevTrans.
    clipRect := nil.

    prevClipArea isNil ifTrue:[device noClipIn:drawableId  gc:gcId]
                      ifFalse:[self clippingRectangle:prevClipArea].

    (w >= (width - margin - margin) and:[h >= (height - margin - margin)]) ifTrue:[
        self sensor flushExposeEventsFor:self.
    ].
! !

!DSVColumnView methodsFor:'scroller interface'!

heightOfContents
    "return the height of the contents in pixels
    "
    ^ self preferredExtent y


!

innerHeight
    "returns the inner height of the contents shown
    "
    ^ height - margin - margin

!

innerWidth
    "returns the inner width of the contents shown
    "
    ^ width - margin - margin

!

verticalScrollStep
    "return the amount to scroll when stepping up/down.
    "
    ^ rowHeight



!

viewOrigin
    "return the viewOrigin; thats the coordinate of the contents 
     which is shown topLeft in the view.
    "
    ^ viewOrigin

!

widthOfContents
    "return the width of the contents in pixels
    "
    ^ self preferredExtent x

!

xOriginOfContents
    "return the horizontal origin of the contents in pixels
    "
    ^ viewOrigin x 

!

yOriginOfContents
    "return the vertical origin of the contents in pixels
    "
    ^ viewOrigin y

! !

!DSVColumnView methodsFor:'scrolling'!

halfPageDown
    "scroll down half a page
    "
    self scrollDown:(self innerHeight // 2)


!

halfPageUp
    "scroll up half a page
    "
    self scrollUp:(self innerHeight // 2)

!

scrollTo:anOrigin redraw:doRedraw
    "change origin to have newOrigin be visible at the top-left.
    "
    |x y dX dY newOrg dltOrg h innerWT innerHG|

    shown ifFalse:[
        ^ self
    ].
    [self sensor hasExposeEventFor:nil] whileTrue:[
        self windowGroup processExposeEvents
    ].
    innerWT := self innerWidth.
    innerHG := self innerHeight.

    h := viewOrigin y.

    (y := anOrigin y) > h ifTrue:[              "/ end of contents
        y > (dY := self maxViewOriginY) ifTrue:[
            y := dY max:h
        ]
    ] ifFalse:[
        y := y max:0.
    ].

    (x := anOrigin x) > 0 ifTrue:[
        x := x min:(self widthOfContents - innerWT).
    ].
    x      := x max:0.
    newOrg := (x @ y).
    dltOrg := newOrg - viewOrigin.
    dX     := dltOrg x.
    dY     := dltOrg y.

    (dX == 0 and:[dY == 0]) ifTrue:[
        ^ self
    ].
    self originWillChange.
    viewOrigin := newOrg.

    doRedraw ifFalse:[
        ^ self originChanged:dltOrg
    ].

    dY ~~ 0 ifTrue:[                            "/ SCROLL VERTICAL
        dY := dY abs.

        (dX ~~ 0 or:[innerHG - dY < 20]) ifTrue:[
            self redraw.
        ] ifFalse:[                             "/ COPY VERTICAL
            |y0 y1|

            y0 := y1 := margin + dY.
            h  := innerHG - dY.

            dltOrg y < 0 ifTrue:[y0 := margin. y := y0]
                        ifFalse:[y1 := margin. y := y1 + h].

            self catchExpose.
            self copyFrom:self x:0 y:y0 toX:0 y:y1 width:width height:h async:true.
            self waitForExpose.
            self redrawX:0 y:y width:width height:(innerHG - h).
        ]
    ] ifFalse:[                                 "/ SCROLL HORIZONTAL
        dX := dX abs.

        innerWT - dX < 20 ifTrue:[
            self redraw.
        ] ifFalse:[                             "/ COPY HORIZONTAL
            |x0 x1 w|

            x0 := x1 := dX.
            w  := width - dX - margin.

            dltOrg x < 0 ifTrue:[x0 := 0. x := x0]
                        ifFalse:[x1 := 0. x := w].

            self catchExpose.
            self copyFrom:self x:x0 y:0 toX:x1 y:0 width:w height:height async:true.
            self waitForExpose.
            self redrawX:x y:0 width:(width - w) height:height.
        ]
    ].
    self originChanged:dltOrg.
!

scrollToRowAt:aRowNr colAt:aColNr
    "make row at a row number in column at a column number visible
    "
    |x y l dY dX|

    (    (aRowNr between:1 and:(self numberOfRows))
     and:[aColNr between:0 and:(self numberOfColumns)]
    ) ifFalse:[
        ^ self
    ].

    dY := dX := 0.
    y  := self yVisibleOfRowNr:aRowNr.

    y < margin ifTrue:[
        dY := margin - y.
    ] ifFalse:[
        y := y + rowHeight.
        l := height - margin.
        y > l ifTrue:[dY := l - y]
    ].

    aColNr == 0 ifTrue:[
        dY == 0 ifTrue:[^ self].
        dX := 0.
    ] ifFalse:[
        x  := self xVisibleOfColNr:aColNr.

        x < margin ifTrue:[
            dX := margin - x
        ] ifFalse:[
            x := x + (self columnAt:aColNr) width.
            l := width - margin.
            x > l ifTrue:[dX := l - x]
        ]
    ].

    (dX == 0 and:[dY == 0]) ifFalse:[
        self scrollTo:(viewOrigin - (dX @ dY)).
    ]


!

scrollToSelection
    "make selection visible
    "
    self hasSelection ifTrue:[
        self scrollToRowAt:selectedRowIndex colAt:selectedColIndex
    ]

! !

!DSVColumnView methodsFor:'selection'!

hasSelection
    "returns true if a selection (raw and column) exists
    "
    ^ selectedRowIndex ~~ 0


!

selectColIndex:aColNr
    "select new column; keep row index
    "
    self selectColIndex:aColNr rowIndex:selectedRowIndex
!

selectColIndex:aColNr rowIndex:aRowNr
    "change selection with notification
    "
    |oC oR|

    oC := selectedColIndex.
    oR := selectedRowIndex.

    self setSelectColIndex:aColNr rowIndex:aRowNr.

    (oC ~~ selectedColIndex or:[oR ~~ selectedRowIndex]) ifTrue:[
        self selectionChanged
    ].
!

selectRow:aRow
    "select a row
    "
    |rowNr|

    (aRow notNil and:[list size ~~ 0]) ifTrue:[
        rowNr := list identityIndexOf:aRow
    ] ifFalse:[
        rowNr := 0
    ].
    self selectRowIndex:rowNr.
!

selectRowIndex:aRowNr
    "select new row; keep column index
    "
    self selectColIndex:selectedColIndex rowIndex:aRowNr
!

selectedColIndex
    "returns selected column number or 0
    "
    ^ selectedColIndex
!

selectedColumn
    "returns selected column or nil
    "
    ^ columnDescriptors at:selectedColIndex ifAbsent:nil.
!

selectedRow
    "returns selected row or nil
    "
    ^ self at:selectedRowIndex ifAbsent:nil
!

selectedRowIndex
    "returns selected row number or 0
    "
    ^ selectedRowIndex
!

selectedRowIndexInColumn:aCol
    selectedRowIndex ~~ 0 ifTrue:[
        (selectedColIndex == 0 or:[self selectedColumn == aCol]) ifTrue:[
            ^ selectedRowIndex
        ]
    ].
    ^ 0
!

selectionChanged
    "selection has changed
    "
    self changed:#selection.

    actionBlock notNil ifTrue:[
        actionBlock value:selectedRowIndex
    ]
!

setSelectColIndex:aColNr rowIndex:aRowNr
    "change selection without notification
    "
    |rowNr colNr newCol oldCol oldRow level sensor|

    (aRowNr between:0 and:(self numberOfRows)) ifTrue:[rowNr := aRowNr]
                                              ifFalse:[rowNr := 0].

    colNr := 0.

    (     rowNr ~~ 0
     and:[(aColNr between:1 and:(self numberOfColumns))
     and:[(self columnAt:aColNr) rendererType ~~ #rowSelector]]
    ) ifTrue:[
        newCol := self columnAt:aColNr.

        newCol canSelect ifTrue:[colNr  := aColNr]
                        ifFalse:[newCol := nil]
        
    ].

    (rowNr == selectedRowIndex and:[colNr == selectedColIndex]) ifTrue:[
        ^ self  "/ nothing changed
    ].

    "/ release old selection

    oldCol := selectedColIndex.
    oldRow := selectedRowIndex.
    selectedRowIndex := rowNr.
    selectedColIndex := colNr.

    oldRow ~~ 0 ifTrue:[
        editValue notNil ifTrue:[
            (self columnAt:oldCol) at:oldRow put:editValue value.
            editValue := nil
        ].

        editView notNil ifTrue:[
            editView destroy.
            editView := nil.
        ].
        "/ redraw old selection unselected
        self realized ifTrue:[
            self redrawRowAt:oldRow colAt:oldCol.
        ]        
    ].   

    "/ show new selection

    self realized ifFalse:[^ self ].

    newCol notNil ifTrue:[
        self scrollToRowAt:rowNr colAt:colNr.

        newCol editorType ~~ #None ifTrue:[

            level := 1 + self separatorSize.
            editView := SimpleView extent:((newCol width - level) @ (rowHeight - level)) in:self.
            self updateEditViewOrigin.
            editView viewBackground:hgLgBgColor.

            editValue := newCol editorForRow:(self at:rowNr)
                                          in:editView
                                        with:hgLgFgColor
                                          bg:hgLgBgColor.
            editView realize.
        ] ifFalse:[
            self redrawRowAt:rowNr colAt:colNr
        ]
    ] ifFalse:[
        selectedRowIndex ~~ 0 ifTrue:[
            self redrawRowAt:rowNr colAt:0.
            self scrollToRowAt:rowNr colAt:0.
        ]
    ].

    (sensor := self sensor) notNil ifTrue:[
        [sensor hasExposeEventFor:nil] whileTrue:[
            self windowGroup processExposeEvents
        ]
    ]


! !

!DSVColumnView class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libwidg2/DSVColumnView.st,v 1.9 1997-10-31 10:25:05 ca Exp $'
! !
