"
 COPYRIGHT (c) 1997 by Claus Gittinger / eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"





View subclass:#DSVLabelView
	instanceVariableNames:'columnView'
	classVariableNames:''
	poolDictionaries:''
	category:'Views-DataSet'
!

!DSVLabelView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1997 by Claus Gittinger / eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"




!

documentation
"
    shows the labels assigned to the column descriptions. Used by the
    DataSetView

    [Instance variables:]

        columnView      <DSVColumnView> column view which shows the columns

    [author:]
        Claus Atzkern

    [see also:]
        DSVColumnView
        DataSetColumnSpec
        DataSetColumn
        DataSetView
"



! !

!DSVLabelView methodsFor:'accessing'!

columnDescriptors:colDesc
    "compute height of baseline
    "
    |h label|

    columnView columnDescriptors:colDesc.
    h := 0.

    columnView numberOfColumns ~~ 0 ifTrue:[
        columnView columnsDo:[:aCol|
            (label := aCol label) notNil ifTrue:[
                h := (label heightOn:self) max:h
            ]
        ]
    ].
    h := h + 8.
    self                 corner:( 1.0 @ h ).
    columnView superView origin:( 0.0 @ h ).
    columnView columnDescriptors:colDesc.

! !

!DSVLabelView methodsFor:'change & update'!

columnsSizeChanged
    |h label spv|

    spv := columnView superView.

    columnView numberOfColumns == 0 ifTrue:[
        self corner:( 1.0 @ 0 ).
        spv  origin:( 0.0 @ 0 ).
      ^ self
    ].

    h := 0.
    columnView columnsDo:[:aCol|
        (label := aCol label) notNil ifTrue:[
            h := (label heightOn:self) max:h
        ]
    ].
    h := h + 8.
    self corner:( 1.0 @ h ).
    spv  origin:( 0.0 @ (h + 1) ).
    self invalidate
!

styleChanged

    super viewBackground:(columnView backgroundColor).
    super font:(columnView font).

    realized ifTrue:[
        self invalidate.
    ].

!

update:what with:aPara from:chgObj
    "one of my models changed
    "
    chgObj == columnView ifFalse:[
        ^ self
    ].

    what == #style ifTrue:[
        ^ self styleChanged
    ].

    what == #sizeOfColumns ifTrue:[
        ^ self columnsSizeChanged.
    ].

    self shown ifTrue:[
        what == #sizeOfContents ifTrue:[
            columnView listSizeChanged ifFalse:[
                self invalidate
            ].
            ^ self
        ].

        what == #originOfContents ifTrue:[
            ^ self srollToDeltaX:(aPara x).
        ]
    ].


! !

!DSVLabelView methodsFor:'initialize / release'!

realize
    self bitGravity:#NorthWest.
    super realize.


! !

!DSVLabelView methodsFor:'instance creation'!

columnView:aColumnView
    columnView := aColumnView.
    columnView addDependent:self.

! !

!DSVLabelView methodsFor:'private'!

leftX
    "the left x startPosition in my view coordinate system; dependent on
     the vertical scroler size, margin and left indent of the column view
    "
    |p|

    p := device translatePoint:(0@0) from:(columnView id) to:(self id).
  ^ columnView leftInset + columnView margin + p x. 


! !

!DSVLabelView methodsFor:'redrawing'!

redraw
    "redraw visible area
    "
    self redrawX:0 y:0 width:width height:height.


!

redrawX:x y:y width:w height:h
    "redraw an area
    "
    |x0 x1 xM prevTransform prevClipArea fg bg bgLbl fgLbl textY hSpace dkCol lgCol has3D|

    self shown ifFalse:[
        ^ self
    ].
    x0 := self leftX - (columnView xOriginOfContents).
    xM := x + w.
    fg := columnView foregroundColor.
    bg := columnView backgroundColor.
    textY := (height + font ascent) // 2.
    hSpace := columnView horizontalSpacing.

    self paint:bg.
    self fillRectangleX:x y:y width:w height:(h min:height).
    columnView numberOfColumns == 0 ifTrue:[
        ^ self
    ].

    prevTransform  := transformation.
    prevClipArea   := clipRect.
    clipRect       := nil.
    transformation := WindowingTransformation new.

    has3D := columnView has3Dsepartors.
    dkCol := columnView separatorDarkColor.
    lgCol := columnView separatorLightColor.

    self paint:fg on:bg.

    columnView columnsDo:[:aCol||label c1 c2|
        x1 := x0 + aCol width.

        (x1 >= x and:[x0 <= xM]) ifTrue:[
            c1 := x0 max:0.
            c2 := x1 min:(width - 1).

            device setClipX:c1 y:0 width:(c2 - c1 + 1) height:height
                         in:drawableId gc:gcId.

            bgLbl := aCol backgroundColor.
            fgLbl := aCol foregroundColor.

            bgLbl notNil ifTrue:[
                self paint:bgLbl.
                self fillRectangleX:x0 y:0 width:(aCol width) height:height.
            ] ifFalse:[
                bgLbl := bg
            ].
         "/ draw label

            (label := aCol label) notNil ifTrue:[                
                label isString ifTrue:[
                    aCol canSelect ifTrue:[
                        fgLbl isNil ifTrue:[fgLbl := fg]
                    ] ifFalse:[
                        fgLbl isNil ifTrue:[fgLbl := dkCol]
                    ].
                    self paint:fgLbl on:bg.
                    label displayOn:self x:x0 + hSpace y:textY
                ] ifFalse:[
                    fgLbl isNil ifTrue:[fgLbl := fg].
                    self paint:fgLbl on:bg.
                    label displayOn:self x:x0 + hSpace y:((height - (label heightOn:self)) // 2)
                ].
            ].
            fgLbl isNil ifTrue:[fgLbl := fg].
            self paint:fgLbl on:bg.

          "/ draw separators

            (     aCol showColSeparator
             and:[aCol ~~ columnView lastColumn
              or:[columnView canFit not]]
            ) ifTrue:[
            "/ aCol showColSeparator ifTrue:[
                has3D ifFalse:[                                                 "/ no 3D-Effect
                    self displayLineFromX:x1     y:0 toX:x1     y:height
                ] ifTrue:[                                                      "/ with 3D-Effect
                    self paint:dkCol.
                    self displayLineFromX:x1     y:0 toX:x1     y:height.
                    self displayLineFromX:x1 - 1 y:0 toX:x1 - 1 y:height.

                    self paint:lgCol.
                    self displayLineFromX:x1 - 2 y:0 toX:x1 - 2 y:height.
                    self displayLineFromX:x1 - 3 y:0 toX:x1 - 3 y:height.
                    self paint:fg on:bg.                                        "/ restore color
                ]
            ]
        ].
        x0 := x1.
    ].
    transformation := prevTransform.
    clipRect := nil.

    prevClipArea isNil ifTrue:[device noClipIn:drawableId  gc:gcId]
                      ifFalse:[self clippingRectangle:prevClipArea].

    self displayLineFromX:0 y:(height - 1) toX:x1 y:(height - 1).


!

srollToDeltaX:deltaX
    "scroll for deltaX pixels
    "
    |absX w x|

    (self shown and:[deltaX ~~ 0]) ifFalse:[
        ^ self
    ].
    absX  := deltaX abs.

    (w := width - absX) < 30 ifTrue:[
        ^ self redraw
    ].

    self catchExpose.

    deltaX > 0 ifTrue:[
        self copyFrom:self x:absX y:0 toX:0  y:0 width:w height:height async:true.
        x := w.
    ] ifFalse:[
        self copyFrom:self x:0 y:0 toX:absX y:0 width:w height:height async:true.
        x := 0.
    ].
    self waitForExpose.
    self redrawX:x y:0 width:absX height:height.


! !

!DSVLabelView class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libwidg2/DSVLabelView.st,v 1.4 1997-10-29 07:42:29 ca Exp $'
! !
